// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

part of ducene;

abstract class Analyzer {
  List<CharFilter> charFilters = [];
  Tokenizer tokenizer;
  List<TokenFilter> tokenFilters = [];
  Iterable<String> getTokens(String text) {
    String normalized = text;
    for (var charFilter in charFilters) {
      normalized = charFilter.normalize(normalized);
    }
    Iterable<String> tokens = tokenizer.tokenize(normalized);
    Iterable<String> filteredTokens = tokens;
    for (var tokenFilter in tokenFilters) {
      filteredTokens = tokenFilter.filter(filteredTokens);
    }
    return filteredTokens;
  }
}
abstract class CharFilter {
  String normalize(String text);
}
abstract class Tokenizer {
  Iterable<String> tokenize(String text);
}
abstract class TokenFilter {
  Iterable<String> filter(Iterable<String> tokens);
}

abstract class PatternReplaceCharFilter implements CharFilter {
  Pattern pattern;
  String replace;
  String normalize(String text) => text.replaceAll(pattern, replace);
}
class PatternCharFilter extends PatternReplaceCharFilter {
  Pattern _pattern;
  String _replace;
  PatternCharFilter(this._pattern, this._replace) {
    pattern = _pattern;
    replace = _replace;
  }
}
class IdeographicSpaceCharFilter extends PatternReplaceCharFilter {
  final pattern = new RegExp(r"　");
  final replace = " ";
}
class DigitSeparatorCharFilter implements CharFilter {
  String normalize(String text) {
    String txt = '';
    for (int i = 0; i < text.length; i++) {
      bool found = false;
      if (text[i] == ',') {
        int prevIndex = i - 1;
        if ((i != 0) && (_isDigit(text, prevIndex))) {
          int nextIndex = i + 1;
          if (!(nextIndex > text.length - 1) && (_isDigit(text, nextIndex))) {
            found = true;
          }
        }
      }
      if (!found) txt = txt + text[i];
    }
    return txt;
  }
  bool _isDigit(String text, int index) {
    return StandardTokenizer.isDigit(text.codeUnitAt(index)) != StandardTokenizer.UNKNOWN;
  }
}

class KeywordTokenizer implements Tokenizer {
  Iterable<String> tokenize(String text) => [text];
}
class WhitespaceTokenizer implements Tokenizer {
  final RegExpTokenizer regExpTokenizer
    = new RegExpTokenizer(new RegExp("[ ,.!?:;'\"\r\n]"));
  Iterable<String> tokenize(String text) {
    return regExpTokenizer.tokenize(text);
  }
}
class RegExpTokenizer implements Tokenizer {
  final RegExp _regExp;
  RegExpTokenizer(this._regExp);
  Iterable<String> tokenize(String text) {
    return text.split(_regExp).where((e) => (e as String).length > 0);
  }
}
class NGramTokenizer implements Tokenizer {
  final int _n;
  NGramTokenizer(this._n);
  Iterable<String> tokenize(String text) {
    return new Iterable.generate(text.length, (i) {
      var endIndex = i + _n;
      if (!(endIndex > text.length)) {
        return text.substring(i, endIndex);
      }
    }).where((e) => e != null);
  }
}
class EdgeNGramTokenizer implements Tokenizer {
  final int _n;
  EdgeNGramTokenizer(this._n);
  Iterable<String> tokenize(String text) {
    return new Iterable.generate(_n, (i) {
      var endIndex = i + 1;
      if (!(endIndex > text.length)) {
        return text.substring(0, endIndex);
      }
    }).where((e) => e != null);
  }
}

class LowerCaseFilter implements TokenFilter {
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) => e.toLowerCase());
  }
}
class NGramFilter implements TokenFilter {
  final int _n;
  NGramFilter(this._n);
  Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      var text = tokens.elementAt(i);
      return new Iterable.generate(text.length, (i) {
        var endIndex = i + _n;
        if (!(endIndex > text.length)) {
          return text.substring(i, endIndex);
        }
      }).where((e) => e != null);
    }).where((e) => e != null).expand((x) => x);
  }
}
class ShingleFilter implements TokenFilter {
  final int _n;
  final String _separator;
  ShingleFilter(this._n, this._separator);
  Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      var shingle = _doShingle(tokens, i, _n);
      return (shingle.length == _n) ? shingle.join(_separator) : null;
    }).where((e) => e != null);
  }
  List<String> _doShingle(Iterable<String> tokens, int i, int n) {
    var ret = [];
    for (int x = i; x < i + n; x++) {
      if (!(x > tokens.length - 1)) {
        ret.add(tokens.elementAt(x));
      }
    }
    return ret;
  }
}
abstract class StopFilter implements TokenFilter {
  List<String> stopWords = [];
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => !stopWords.contains(e));
  }
}
class StopWordsFilter extends StopFilter {
  final List<String> _stopWords;
  StopWordsFilter(this._stopWords) {
    stopWords = _stopWords;
  }
}
class EnglishStopWordsFilter extends StopFilter {
  final List<String> stopWords = [
    "a", "an", "and", "are", "as", "at", "be", "but", "by",
    "for", "if", "in", "into", "is", "it",
    "no", "not", "of", "on", "or", "such",
    "that", "the", "their", "then", "there", "these",
    "they", "this", "to", "was", "will", "with"];
}
class JapaneseStopWordsFilter extends StopFilter {
  final List<String> stopWords = [
    'の','に','は','を','た','が','で','て','と','し','れ','さ',
    'ある','いる','も','する','から','な','こと','として',
    'い','や','れる','など','なっ','ない','この','ため','その',
    'あっ','よう','また','もの','という','あり','まで','られ',
    'なる','へ','か','だ','これ','によって','により','おり','より',
    'による','ず','なり','られる','において','ば','なかっ','なく',
    'しかし','について','せ','だっ','その後','できる','それ','う',
    'ので','なお','のみ','でき','き','つ','における','および','いう',
    'さらに','でも','ら','たり','その他','に関する','たち',
    'ます','ん','なら','に対して','特に','せる','及び',
    'これら','とき','では','にて','ほか','ながら','うち',
    'そして','とともに','ただし','かつて','それぞれ','または',
    'お','ほど','ものの','に対する','ほとんど','と共に','といった',
    'です','とも','ところ','ここ'];
}
abstract class KeepFilter implements TokenFilter {
  List<String> keepWords = [];
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => keepWords.contains(e));
  }
}
class KeepWordsFilter extends KeepFilter {
  final List<String> _keepWords;
  KeepWordsFilter(this._keepWords) {
    keepWords = _keepWords;
  }
}
abstract class OrderingFilter implements TokenFilter {
  bool reverse = false;
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) {
      List<String> str = new Iterable.generate(e.length, (i) => e[i]).toList();
      List<String> ordered = [];
      str.sort();
      ordered.addAll(str);
      if (reverse) {
        ordered = ordered.reversed.toList();
      }
      return ordered.join();
    });
  }
}
class OrderingNaturalFilter extends OrderingFilter {
  final reverse = false;
}
class OrderingReverseFilter extends OrderingFilter {
  final reverse = true;
}
class ReverseStringFilter implements TokenFilter {
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) {
      List<String> str = new Iterable.generate(e.length, (i) => e[i]).toList();
      List<String> ordered = str.reversed.toList();
      return ordered.join();
    });
  }
}
class LengthFilter implements TokenFilter {
  final int _min;
  final int _max;
  LengthFilter(this._min, this._max);
  Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => (_min <= e.length && e.length <= _max));
  }
}
class TypoToleranceFilter implements TokenFilter {
  final int _n;
  NGramFilter _nGram;
  final List<OrderingFilter> _orders
    = [new OrderingNaturalFilter(), new OrderingReverseFilter()];
  TypoToleranceFilter(this._n) {
    _nGram = new NGramFilter(_n);
  }
  Iterable<String> filter(Iterable<String> tokens) {
    var ite = new Iterable.generate(_orders.length, (i) {
      var order = _orders[i];
      return order.filter(tokens);
    });
    return _nGram.filter(ite.expand((x) => x));
  }
}

class KeywordAnalyzer extends Analyzer {
  final tokenizer = new KeywordTokenizer();
}
class LowerCaseAnalyser extends Analyzer {
  final charFilters = [new IdeographicSpaceCharFilter()];
  final tokenizer = new KeywordTokenizer();
  final tokenFilters = [new LowerCaseFilter()];
}
class WhitespaceAnalyzer extends Analyzer {
  final tokenizer = new WhitespaceTokenizer();
  final tokenFilters = [new LowerCaseFilter()];
}
class RegExpAnalyzer extends Analyzer {
  final RegExp _regExp;
  RegExpAnalyzer(this._regExp) {
    tokenizer = new RegExpTokenizer(_regExp);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class NGramAnalyzer extends Analyzer {
  final int _n;
  NGramAnalyzer(this._n) {
    tokenizer = new NGramTokenizer(_n);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class EdgeNGramAnalyzer extends Analyzer {
  final int _n;
  EdgeNGramAnalyzer(this._n){
    tokenizer = new EdgeNGramTokenizer(_n);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class ShingleAnalyzer extends Analyzer {
  final Tokenizer _tokenizer;
  final List<TokenFilter> _tokenFilters;
  final int _n;
  final String _separator;
  ShingleAnalyzer(this._tokenizer, this._tokenFilters, this._n, this._separator) {
    tokenizer = _tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_tokenFilters);
    filters.add(new ShingleFilter(_n, _separator));
    tokenFilters = filters;
  }
}
class StandardAnalyzer extends Analyzer {
  final charFilters = [new DigitSeparatorCharFilter()];
  final tokenizer = new StandardTokenizer();
  final tokenFilters
    = [new LowerCaseFilter(),
       new EnglishStopWordsFilter(), new JapaneseStopWordsFilter()];
}
class PhraseAnalyzer extends Analyzer {
  final Analyzer _analyzer;
  final String _separator;
  PhraseAnalyzer(this._analyzer, this._separator) {
    charFilters = _analyzer.charFilters;
    tokenizer = _analyzer.tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_analyzer.tokenFilters);
    filters.add(new ShingleFilter(2, _separator));
    tokenFilters = filters;
  }
}
class TypoToleranceAnalyzer extends Analyzer {
  final Analyzer _analyzer;
  TypoToleranceAnalyzer(this._analyzer) {
    charFilters = _analyzer.charFilters;
    tokenizer = _analyzer.tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_analyzer.tokenFilters);
    filters.add(new TypoToleranceFilter(4));
    tokenFilters = filters;
  }
}
class UniGramAnalyzer extends Analyzer {
  final tokenizer = new NGramTokenizer(1);
}
class BiGramAnalyzer extends Analyzer {
  final tokenizer = new NGramTokenizer(2);
}

class StandardTokenizer implements Tokenizer {
  static int isDigit(int code) {
    return (0x0030 <= code && code <= 0x0039) ? 1 : UNKNOWN;
  }
  int _isLatin(int code) {
    return ((0x0041 <= code && code <= 0x005A) || (0x0061 <= code && code <= 0x007A)) ? 2 : UNKNOWN;
  }
  bool _ideographic(int code) => (0x4E00 <= code && code <= 0x9FD5);
  bool _ideographicA(int code) => (0x3400 <= code) && (code <= 0x4DFF);
  bool _ideographicB(int code) => (0x20000 <= code) && (code <= 0x2A6FF);
  bool _ideographicC(int code) => (0x2A700 <= code) && (code <= 0x2B734);
  bool _ideographicD(int code) => (0x2B740 <= code) && (code <= 0x2B81F);
  bool _ideographicE(int code) => (0x2B820 <= code) && (code <= 0x2CEAF);
  int _isIdeographic(int code) {
    if (_ideographic(code)
        || _ideographicA(code) || _ideographicB(code) || _ideographicC(code)
        || _ideographicD(code) || _ideographicE(code)) {
      return 3;
    }
    return UNKNOWN;
  }
  int _isHiragana(int code) {
    return (0x3041 <= code && code <= 0x3096) ?  4 : UNKNOWN;
  }
  int _isKatakana(int code) {
    return (0x30A1 <= code && code <= 0x30FA) || (code == 0x30FC) ? 5 : UNKNOWN;
  }
  int _charType(int code) {
    int a = isDigit(code);
    if (a != UNKNOWN) return a;
    int b = _isLatin(code);
    if (b != UNKNOWN) return b;
    int c = _isIdeographic(code);
    if (c != UNKNOWN) return c;
    int d = _isHiragana(code);
    if (d != UNKNOWN) return d;
    int e = _isKatakana(code);
    if (e != UNKNOWN) return e;
    return UNKNOWN;
  }
  static int UNKNOWN = -1;
  static int _INIT_STATE = -10000;
  int _previousState = _INIT_STATE;
  void _updateState(int type) {
    _previousState = type;
  }
  bool _shouldAdd(int type) => (type == UNKNOWN) ? false : true;
  Iterable<String> tokenize(String text) {
    var ret = [];
    var tmpChars = [];
    for (int i = 0; i < text.length; i++) {
      int code = text.codeUnitAt(i);
      int type = _charType(code);
      int prevState = _previousState; _updateState(type);
      if (_shouldAdd(type)) {
        if (prevState != type) _emit(tmpChars, ret); // should break by type
        tmpChars.add(text[i]);
      } else {
        _emit(tmpChars, ret);
      }
    }
    _emit(tmpChars, ret);
    return ret;
  }
  void _emit(List tmpChars, List ret) {
    if (!tmpChars.isEmpty) {
      String str = tmpChars.join();
      ret.add(str);
      tmpChars.clear();
    }
  }
}